﻿Imports Brain.Backpropagation
''' <summary>
''' This class uses simplified picture representation given to a neural net to recognice a letter.
''' </summary>
Public Class BasicOCR
    ''' <summary>
    ''' Combinaion of a letter and its simplified binary representation.
    ''' </summary>
    <Serializable()> _
     Private Structure TrainingSet
        Dim cLetter As Char
        Dim bRepr As Boolean(,)
    End Structure
    ''' <summary>
    ''' Combination of letter and a parameter how sure the network answer is.
    ''' </summary>
    Public Structure RecognitionResult
        Dim cLetter As Char
        Dim dQuality As Double
    End Structure
    ''' <summary>
    ''' Defines in which order the training sets will be trained into the network.
    ''' </summary>
    Public Enum TrainingMode
        ''' <summary>
        ''' Usually better, but slower.
        ''' </summary>
        Randomly
        ''' <summary>
        ''' Faster, but the network may learn some correlation from dataset order.
        ''' </summary>
        Linear
    End Enum
    Private Shared brain As New Brain.Backpropagation(25, 26, 40, 0.5, 0.1)
    Private Shared trainings As New List(Of TrainingSet)
    ''' <summary>
    ''' Adds the given combination to training set list.
    ''' Returns the quantity of training sets available.
    ''' </summary>
    Public Shared Function addTrainingSet(ByVal bRepr As Boolean(,), ByVal cLetter As Char) As Integer
        Dim ts As New TrainingSet
        ts.cLetter = cLetter
        ts.bRepr = bRepr
        trainings.Add(ts)
        Return trainings.Count
    End Function
    ''' <summary>
    ''' Trains all training sets to the neural net.
    ''' This may take a while.
    ''' </summary>
    Public Shared Sub trainTrainingSet(ByVal dMaxError As Double, ByVal mode As TrainingMode)
        Dim dInputs As New List(Of Double())
        Dim dOutputs As New List(Of Double())
        For Each ts In trainings
            dInputs.Add(getNetInputFromRepresentation(ts.bRepr).ToArray)
            dOutputs.Add(getNetOutputFromLetter(ts.cLetter).ToArray)
        Next
        Select Case mode
            Case TrainingMode.Linear
                brain.trainLinear(dInputs, dOutputs, dMaxError)
            Case TrainingMode.Randomly
                brain.trainRandom(dInputs, dOutputs, dMaxError)
        End Select
    End Sub
    ''' <summary>
    ''' Saves all currently available trainging sets to file.
    ''' </summary>
    Public Shared Sub saveTrainingSets(ByVal sFileName As String)
        Try
            If My.Computer.FileSystem.FileExists(sFileName) Then
                My.Computer.FileSystem.DeleteFile(sFileName)
            End If
            Dim fs As IO.FileStream = New IO.FileStream(sFileName, IO.FileMode.OpenOrCreate)
            Dim bf As New Runtime.Serialization.Formatters.Binary.BinaryFormatter()
            bf.Serialize(fs, trainings)
            fs.Close()
        Catch ex As Exception
            MsgBox(ex.Message, MsgBoxStyle.Exclamation)
        End Try
    End Sub
    ''' <summary>
    ''' Loads all training sets from file.
    ''' Returns the quanitity of available training sets.
    ''' </summary>
    ''' <param name="sFileName">File path where to read.</param>
    Public Shared Function loadTrainingSets(ByVal sFileName As String) As Integer
        Dim iLoaded As Integer = 0
        Try
            If My.Computer.FileSystem.FileExists(sFileName) Then
                Dim fsRead As New IO.FileStream(sFileName, IO.FileMode.Open)
                Dim bf As New Runtime.Serialization.Formatters.Binary.BinaryFormatter()
                trainings = CType(bf.Deserialize(fsRead), List(Of TrainingSet))
                fsRead.Close()
                iLoaded = trainings.Count
            End If
        Catch ex As Exception
            MsgBox("Could not load training sets: " & ex.Message, MsgBoxStyle.Exclamation)
        End Try
        Return iLoaded
    End Function
    ''' <summary>
    ''' Asks neural net for the letter represended by boolean field.
    ''' </summary>
    Public Shared Function getLetter(ByVal bRepr As Boolean(,)) As Char
        Return getResult(bRepr).cLetter
    End Function
    ''' <summary>
    ''' Returns the recognition result to given input representation.
    ''' </summary>
    Public Shared Function getResult(ByVal bRepr As Boolean(,)) As RecognitionResult
        Dim iIndexMax As Integer = 0
        Dim dQuality As Double = 0
        Dim dResult() As Double = brain.calc(getNetInputFromRepresentation(bRepr).ToArray)
        For i = 0 To UBound(dResult)
            If dResult(i) > dQuality Then
                dQuality = dResult(i)
                iIndexMax = i
            End If
        Next
        Dim rr As New RecognitionResult
        rr.cLetter = getLetterFromIndex(iIndexMax)
        rr.dQuality = dQuality
        Return rr
    End Function
    ''' <summary>
    ''' Returns a 0-based index of lower case alphabet letters.
    ''' </summary>
    Private Shared Function getLetterIndex(ByVal cLetter As Char) As Integer
        Return CInt(AscW(cLetter)) - 97
    End Function
    ''' <summary>
    ''' Return the char from 0-based index of lower case alphabet letters.
    ''' </summary>
    Private Shared Function getLetterFromIndex(ByVal iIndex As Integer) As Char
        Return Chr(iIndex + 97)
    End Function
    ''' <summary>
    ''' Converts the binary representation in linear net input.
    ''' </summary>
    Private Shared Function getNetInputFromRepresentation(ByRef bRepr As Boolean(,)) As List(Of Double)
        Dim dInput As New List(Of Double)
        For iX = 0 To bRepr.GetUpperBound(0)
            For iY = 0 To bRepr.GetUpperBound(1)
                If bRepr(iX, iY) = True Then
                    dInput.Add(0.5)
                Else
                    dInput.Add(-0.5)
                End If
            Next
        Next
        Return dInput
    End Function
    ''' <summary>
    ''' Converts a char into neural net output code.
    ''' </summary>
    Private Shared Function getNetOutputFromLetter(ByVal cLetter As Char) As List(Of Double)
        Dim dOutput As New List(Of Double)
        Dim iChar As Integer = getLetterIndex(cLetter)
        For i = 0 To 25
            If i = iChar Then
                dOutput.Add(1.0)
            Else
                dOutput.Add(0.0)
            End If
        Next
        Return dOutput
    End Function
End Class
